# -*- coding: utf-8 -*-
"""
Created on Mon Feb  1 13:38:24 2016

@author: benny
"""

from __future__ import print_function

from copy import copy
from .ode import ode

def odeint(rhsfun, tout, y0,  method='bdf', **options):
    """
    Integrate a system of ordinary differential equations.
    *odeint* is a wrapper around the ode class, as a convenience function to
    quickly integrate a system of ode.
    Solves the initial value problem for stiff or non-stiff systems
    of first order ode's:

        rhs = dy/dt = fun(t, y)

    where y can be a vector, then rhsfun must be a function computing rhs with
    signature:

        rhsfun(t, y, rhs)

    storing the computed dy/dt in the rhs array passed to the function

    Parameters
    ----------
    rhsfun : callable(t, y, out)
        Computes the derivative at dy/dt in terms of t and y, and stores in out
    y0 : array
        Initial condition on y (can be a vector).
    t : array
        A sequence of time points for which to solve for y.  The initial
        value point should be the first element of this sequence.
    method : string, solution method to use.
        Available are all the ode class solvers as well as some convenience
        shorthands:

        =======  ==============================================================
        Method   Meaning
        =======  ==============================================================
        bdf      This uses the 'cvode' solver in default from, which is a
                 variable step, variable coefficient Backward Differentiation
                 Formula solver, good for stiff ODE. Newton iterations are
                 used to solve the nonlinear system.
        admo     This uses the 'cvode' solver with option lmm_type='ADAMS',
                 which is a variable step Adams-Moulton method (linear
                 multistep method), good for non-stiff ODE. Functional
                 iterations are used to solve the nonlinear system.
        rk5      This uses the 'dopri5' solver, which is a variable step
                 Runge-Kutta method of order (4)5 (use for non-stiff ODE)
        rk8      This uses the 'dop853' solver, which is a variable step
                 Runge-Kutta method of order 8(5,3)
        =======  ==============================================================

        For educational purposes, you can also access following methods:

        =======  ==============================================================
        Method   Meaning
        =======  ==============================================================
        beuler   This is the Implicit Euler (backward Euler) method (order 1),
                 which is obtained via the 'bdf' method, setting the order
                 option to 1, setting large tolerances,  and fixing the
                 stepsize.
                 Use option 'step' to change stepsize, default: step=0.05.
                 Use option 'rtol' and 'atol' to use more strict tolerances
                 Note: this is not completely the backward Euler method, as
                 the cvode solver has added control options!
        trapz    This is the Trapezoidal Rule method (order 2), which is
                 obtained via the 'admo' method, setting option order to 2,
                 setting large tolerances and fixing the stepsize.
                 Use option 'step' to change stepsize, default: step=0.05.
                 Use option 'rtol' and 'atol' to use more strict tolerances
                 Note: The cvode solver might change the order to 1 internally
                 in which case this becomes beuler method. Set atol, rtol
                 options as strict as possible.
        =======  ==============================================================

        You can also access the solvers of ode via their names:

        =======  ==============================================================
        Method   Meaning
        =======  ==============================================================
        cvode    This uses the 'cvode' solver
        dopri5   This uses the 'dopri5' solver
        dop853   This uses the 'dop853' solver
        =======  ==============================================================

    options : extra solver options, optional
        Every solver has it's own extra options, see the ode class and the
        details of the solvers available there to know the options possible per
        solver

    Returns
    -------
    solution : named tuple
        A single named tuple is returned containing the result of the
        integration.

        ========  ==========================================
        Field     Meaning
        ========  ==========================================
        flag      An integer flag
        values    Named tuple with fields t and y
        errors    Named tuple with fields t and y
        roots     Named tuple with fields t and y
        tstop     Named tuple with fields t and y
        message   String with message in case of an error
        ========  ==========================================

    See Also
    --------
    scikits.odes.ode.ode : a more object-oriented integrator
    scikits.odes.dae.dae : a solver for differential-algebraic equations
    scipy.integrate.quad : for finding the area under a curve

    Examples
    --------
    The second order differential equation for the angle `theta` of a
    pendulum acted on by gravity with friction can be written:

    .. math:: \\theta''(t) + b \\theta'(t) + c \\sin(\\theta(t)) = 0

    where `b` and `c` are positive constants, and a prime (') denotes a
    derivative.  To solve this equation with `odeint`, we must first convert
    it to a system of first order equations.  By defining the angular
    velocity ``omega(t) = theta'(t)``, we obtain the system:

    .. math:: \\theta'(t) = \\omega(t)
     \\omega'(t) = -b \\omega(t) - c \\sin(\\theta(t))

    We assume the constants are `b` = 0.25 and `c` = 5.0:

    >>> b = 0.25
    >>> c = 5.0

    Let `y` be the vector [`theta`, `omega`].  We implement this system
    in python as:

    >>> def pend(t, y, out):
     ...     theta, omega = y
     ...     out[:] = [omega, -b*omega - c*np.sin(theta)]
     ...

    In case you want b and c easily changable, make pend a class method, and
    consider attributes b and c accessible via `self.b` and `self.c`.
    For initial conditions, we assume the pendulum is nearly vertical
    with `theta(0)` = `pi` - 0.1, and it initially at rest, so
    `omega(0)` = 0.  Then the vector of initial conditions is

    >>> y0 = [np.pi - 0.1, 0.0]

    We generate a solution 101 evenly spaced samples in the interval
    0 <= `t` <= 10.  So our array of times is

    >>> t = np.linspace(0, 10, 101)

    Call `odeint` to generate the solution.

    >>> from scikits.odes.odeint import odeint
    >>> sol = odeint(pend, t, y0)

    The solution is a named tuple `sol`. sol.values.y is an array with shape (101, 2).
    The first column is `theta(t)`, and the second is `omega(t)`.
    The following code plots both components.

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(sol.values.t, sol.values.y[:, 0], 'b', label='theta(t)')
    >>> plt.plot(sol.values.t, sol.values.y[:, 1], 'g', label='omega(t)')
    >>> plt.legend(loc='best')
    >>> plt.xlabel('t')
    >>> plt.grid()
    >>> plt.show()
    """
    t = copy(tout)
    y0 = copy(y0)

    int_name = 'cvode'
    #always use old api
    options['old_api'] = False
    if (method == 'bdf'):
        int_name = 'cvode'
    elif (method == 'admo'):
        options['lmm_type'] = 'ADAMS'
        options['nonlinsolver'] = 'fixedpoint'
        int_name = 'cvode'
    elif (method == 'rk5'):
        int_name = 'dopri5'
    elif (method == 'rk8'):
        int_name = 'dop853'
    elif (method == 'beuler' or method == 'trapz'):
        int_name = 'cvode'
        options['order'] = 1
        if not options.has_key('step'):
            options['step'] = 0.05
        options['min_step_size'] = options['step']
        options['max_step_size'] = options['step']
        #no error control possible, try to disable it:
        if not options.has_key('atol'):
            options['atol'] = 1e3
        if not options.has_key('rtol'):
            options['rtol'] = 1e3
        del options['step']
        if (method == 'trapz'):
            options['lmm_type'] = 'ADAMS'
            options['nonlinsolver'] = 'fixedpoint'
            options['order'] = 2
    else:
        int_name = method

    solver = ode(int_name, rhsfun, **options)
    return solver.solve(t, y0)



# -*- coding: utf-8 -*-
# Authors: B. Malengier 
"""
lsodi
=====
Solver developed during the 1980s, this is the version from 1987 - Fortran

Integrator for linearly implicit systems of first-order odes.  lsodi
provides the same methods as vode (adams and bdf).  This integrator
accepts the following parameters in initializer or set_options method of the 
dae class:

- rfn : residual function, see below for signature. This option need not be
        set, as rfn will be set during initialization. If the residual function
        of initialization needs to be reset, this option can be used
- jacfn : jacobian function, see below for signature. Default is None.
- adda_func = function, required, see below
- atol=float|seq
- rtol=float
- lband=None|int
- rband=None|int
- method='adams'|'bdf'
- with_jacobian=0|1
- max_steps = int
- max_step_size = float
- (first|min|max)_step = float
- tstop = None|float
- order = int        # <=12 for adams, <=5 for bdf
- compute_initcond = None|'yp0'

Details: 

Implicit integration requires two functions do be defined to integrate 
                    d y(t)[i]
           a(t,y) * --------- = g(t,y)[i]
                        dt
where a(t,y) is a square matrix.

- rfn computes the residual which the user provides.
  rfn has the signature: rfn(x, y, yprime, return_residual)
with 
    x       : independent variable, eg the time, float
    y       : array of n unknowns in x
    yprime  : dy/dx array of n unknowns in x, dimension = dim(y)
    return_residual: array that must be updated with the value of the residuals,
              so G(t,y,y').  The dimension is equal to dim(y)
    return value: Not needed. However, for use with other solvers, consider 
              returning 0 on success.
It calculates something like
        def res(t, y, yprime, r):
            r(t,y,s)=g(t,y)-a(t,y)*yprime

- adda_func must modify the provided matrix p and is a required option.  If banded storage is used,
ml and mu provide the upper and lower diagonals, see the lsodi.f
source for full documentation.  adda_func is passed to the dae via set_options
with the adda_func keyword argument. Schematically, it is
        def adda(t, y, ml, mu, p, nrowp):
            p += a(t,y)
            return p
Note: if your residual is a yprime - g, then adda must substract a, see def of
rfn !

- jacfn is not a supported option for lsodi.

- compute_initcond: None or 'yp0'
  LSODI may be able to compute the initial conditions if you do not know them
  precisely and the problem is not algebraic at t=0. 
  If yp0, then the differential variables (y of the ode system at time 0) 
    will be used to solve for the derivatives of the differential variables,
    so yp0 will be calculated. 
Source: http://www.netlib.org/ode/lsodi.f

"""

from __future__ import print_function

__all__ = ['lsodi']
__version__ = "$Id$"
__docformat__ = "restructuredtext en"

from numpy import asarray, array, zeros, sin, int32, isscalar, empty, alen
from copy import copy
from .dae import DaeBase
import re, sys

class lsodi(DaeBase):
    try:
        from .lsodi import lsodi as _runner
        from .lsodi import intdy as _intdy
    except ImportError:
        print(sys.exc_info()[1])
        _runner = None
        _intdy = None

    name = 'lsodi'

    messages = {2 : 'lsodi was successful.',
               -1 : 'excess work done on this call (check all inputs).',
               -2 : 'excess accuracy requested (tolerances too small).',
               -3 : 'illegal input detected (see printed message).',
               -4 : 'repeated error test failures (check all inputs).',
               -5 : 'repeated convergence failures (perhaps bad jacobian'
                    ' supplied or wrong choice of tolerances).',
               -6 : 'error weight became zero during problem. (solution'
                    ' component i vanished, and atol or atol(i) = 0.).',
               -7 : 'cannot occur in casual use.',
               -8 : 'lsodi was unable to compute the initial dy/dt.  In'
                    ' casual use, this means a(t,y) is initially singular.'
                    '  Supply ydoti and use istate = 1 on the first call.'
               }
    supports_run_relax = 1
    supports_step = 1

    def __init__(self, resfn, **options):
        default_values = {
                'rtol': 1e-6,
                'atol': 1e-12,
                'lband': None,
                'uband': None,
                'tstop': None, 
                'order': 0,
                'max_steps': 500,
                'max_step_size': 0.0, #corresponds to infinite
                'min_step': 0.0,
                'first_step': 0.0, # determined by solver
                'method': "adams", 
                'compute_initcond': None,
                'adda_func': None,
                'rfn': None,
                'jacfn': None,
                 }
        self.t = None
        self.y = None
        self.yp = None
        self.tmp_res = None
        self.tmp_jac = None
        self.options = default_values
        self.set_options(rfn=resfn, **options)
        self.initialized = False

    def set_options(self, **options):
        """ See dae.DaeBase
        """
        for (key, value) in options.items():
            self.options[key.lower()] = value
        self.initialized = False

    def _init_data(self):
        self.rtol = self.options['rtol']
        self.atol = self.options['atol']
        self.mu = self.options['uband']
        self.ml = self.options['lband']
        self.jac = self.options['jacfn']
        if self.jac is not None:
            raise ValueError('Using provided Jacobian is not supported by lsodi')
        self.res = self.options['rfn']

        self.tstop = self.options['tstop']
        self.order = self.options['order']
        self.nsteps = self.options['max_steps']
        self.max_step = self.options['max_step_size']
        self.min_step = self.options['min_step']
        self.first_step = self.options['first_step']
        if re.match(self.options['method'], r'adams', re.I): self.meth = 1
        elif re.match(self.options['method'], r'bdf', re.I): self.meth = 2
        else: raise ValueError('Unknown integration method %s'%(self.options['method']))
        if self.options['compute_initcond'] is None:
            self.compute_initcond = 0
        elif re.match(self.options['compute_initcond'], r'yp0', re.I):
            self.compute_initcond = 1
        else:
            raise ValueError('Unknown init cond calculation method %s' %(
                                            self.options['compute_initcond']))
                                            

        if self.options['adda_func'] is None:
            raise ValueError('adda_func is required for lsodi algorithm!')
        self.adda = self.options['adda_func']
        self.success = 1

    def init_step(self, t0, y0, yp0, y_ic0_retn = None, yp_ic0_retn = None):
        """ See dae.DaeBase
        """
        self._init_data()
        self.y0 = y0
        self.yp0 = yp0
        self.t = t0
        self._reset(len(y0), self.jac is not None)
        
        if self.compute_initcond:
            if self.first_step <= 0.:
                first_step = 1e-18
            else:
                first_step = self.first_step
            self.y, self.yp, t = self.__run(y0, yp0, t0, t0 + first_step)
            t0_init = t
        else:
            self.y = copy(y0)
            self.yp = copy(yp0)
            t0_init = t0

        if not y_ic0_retn is None: y_ic0_retn[:] = self.y[:]
        if not yp_ic0_retn is None: yp_ic0_retn[:] = self.yp[:]
        self.initialized = True
        return t0_init

    def _reset(self, n, has_jac):
        # Calculate parameters for Fortran subroutine lsodi.
        self.neq = n
        if has_jac:
            if self.mu is None and self.ml is None:
                miter = 1
            else:
                if self.mu is None: self.mu = 0
                if self.ml is None: self.ml = 0
                miter = 4
        else:
            if self.mu is None and self.ml is None:
                miter = 2
            else:
                if self.mu is None: self.mu = 0
                if self.ml is None: self.ml = 0
                miter = 5
        mf = 10*self.meth + miter
        if mf in [11,12]:
            lrw = 22 + 16*n + n**2
        elif mf in [14,15]:
            lrw = 22 + 17*n + (2*ml +mu)*neq
        elif mf in [21,22]:
            lrw = 22 + 9*n + n**2
        elif mf in [24,25]:
            lrw = 22 + 10*n + (2*ml +mu)*neq
        else:
            raise ValueError('Unexpected mf=%s'%(mf))
        liw = 20 + n
        rwork = zeros((lrw,), float)
        rwork[4] = self.first_step
        rwork[5] = self.max_step
        rwork[6] = self.min_step
        self.rwork = rwork
        iwork = zeros((liw,), int32)
        if self.ml is not None:
            iwork[0] = self.ml
        if self.mu is not None:
            iwork[1] = self.mu
        iwork[4] = self.order
        iwork[5] = self.nsteps
        iwork[6] = 2           # mxhnil
        self.iwork = iwork
        if isscalar(self.atol) :
            itol=1
        else:
            itol=2
        itask=1
        # don't internally compute ydot if it is provided
        if self.compute_initcond==0:
            istate=1
        else:
            istate=0
        if self.tstop is not None:
            self.rwork[0]=self.tstop
        iopt=1
        self.call_args = [itol,self.rtol,self.atol,itask,istate,iopt,
                            self.rwork,self.iwork,mf]
        #create storage
        self.tmp_res = empty(self.neq, float)
        if (self.ml is None or self.mu is None) :
            self.tmp_jac = zeros((self.neq, self.neq), float)
        else:
            self.tmp_jac = zeros((self.ml + self.mu + 1, self.neq), float)
        self.success = 1

    def _resFn(self, t, y, yp):
        """Wrapper around the residual as defined by user, to create
         the residual needed by lsodi
        """
        self.res(t, y, yp, self.tmp_res)
        return self.tmp_res

    def _jacFn(self, t, y, yp, cj):
        """Wrapper around the jacobian as defined by user, to create
         the jacobian needed by lsodi
        """
        ## Not supported for lsodi
        #self.jac( t, y, yp, cj, self.tmp_jac)
        return self.tmp_jac

    def solve(self, tspan, y0, yp0, hook_fn = None):
        """ See dae.DaeBase
        """
        
        t_retn = empty([alen(tspan), ], float)
        y_retn = empty([alen(tspan), alen(y0)], float)
        yp_retn = empty([alen(tspan), alen(y0)], float)
        
        y_ic0_retn = empty(alen(y0), float)
        yp_ic0_retn = empty(alen(y0), float)
        tinit = self.init_step(tspan[0], y0, yp0, y_ic0_retn, yp_ic0_retn)
        
        t_retn[0] = tinit
        y_retn[0,:] = y0[:]
        yp_retn[0, :] = yp0[:]
        if self.tstop is None:
            itask = 1
            self.call_args[3] = 1
        else:
            itask = self.call_args[3]
            self.call_args[3] = 4
        indbreak = None
        for ind, time in enumerate(tspan[1:]):
            result = self.__run(y_retn[ind], yp_retn[ind], t_retn[ind], time)
            if not self.success:
                indbreak = ind+1
                break
            t_retn[ind+1] = result[2]
            y_retn[ind+1][:] = result[0][:]
            yp_retn[ind+1][:] = result[1][:]
        self.t = t_retn[-1]
        if indbreak is not None:
            self.t = t_retn[indbreak-1]
            return self.success, t_retn[:indbreak], y_retn[:indbreak],\
                   yp_retn[:indbreak], t_retn[indbreak], y_retn[indbreak],\
                   yp_retn[indbreak]
        return self.success, t_retn, y_retn, yp_retn, None, None, None

    def __run(self, *args):
        y1, y1prime_tmp, t, istate = self._runner(*((self._resFn, self.adda,
                                                 self._jacFn) + args[0:]
                                                 + tuple(self.call_args)) 
                                             )
        self.call_args[4] = istate
        y1prime = None
        if istate <0:
            print('lsodi:',self.messages.get(istate,'Unexpected istate=%s'%istate))
            if istate in [-1,-4,-5] :
                print('lsodi: present residual is', y1prime_tmp)
            self.success = 0
            
        if self.success:
            yh = self.rwork[20:]
            order = 1
            y1prime, iflag = self._intdy(t, order, yh, self.neq)
            if iflag<0:
                if iflag==-1: 
                    raise ValueError("order=%s invalid in call to intdy" \
                                                        %order)
                if iflag==-2: 
                    raise ValueError("t=%s invalid in call to intdy"%t)
        return y1, y1prime, t

    def step(self, t, y_retn, yp_retn = None):
        """ See dae.DaeBase
        """
        if not self.initialized:
            raise ValueError('Method ''init_step'' has to be called prior to the'
                    'first call of ''step'' method, or after changing options')
        if t > 0.0:
            if self.tstop is None:
                itask = 1
                self.call_args[3] = 1
            else:
                itask = self.call_args[3]
                self.call_args[3] = 4
            self.y, self.yp, self.t = self.__run(self.y, self.yp, self.t, t)
            self.call_args[3] = itask
        else:
            itask = self.call_args[3]
            if self.tstop is None:
                self.call_args[3] = 2
            else:
                self.call_args[3] = 5
            self.y, self.yp, self.t  = self.__run(self.y, self.yp, self.t, -t)
            self.call_args[3] = itask
        y_retn[:] = self.y[:]
        if yp_retn is not None:
            yp_retn[:] = self.yp[:]
        return self.call_args[4], self.t

if lsodi._runner:
    DaeBase.integrator_classes.append(lsodi)



# Authors: B. Malengier, russel (scipy trac)
from __future__ import print_function

"""
Tests for differential algebraic equation solvers.
Here we test onroot and ontstop
"""
import numpy as np

from numpy import (arange, zeros, array, dot, sqrt, cos, sin, allclose,
                    empty, alen)

from numpy.testing import TestCase, run_module_suite

from scikits.odes import ode
from scikits.odes.sundials.cvode import StatusEnum
from scikits.odes.sundials.common_defs import DTYPE

#data
g  = 9.81    # gravitational constant

Y0 = 1000.0  # Initial height
Y1 = 10.0    # Bottom height - when reached, changes to Y0 (teleport)
T1 = 10.0    # stop time
v0 = 0.0     # Initial speed

#initial data at t=0, y[0] = Y, y[1] = \dot{Y}
y0 = [Y0, v0]
t_end1 =  10.0 # Time of free fall for experiments 1,2
t_end2 = 100.0 # Time of free fall for experiments 3,4

atol = 1e-4
rtol = 1e-4

def rhs_fn(t, y, ydot):
    """ rhs equations for the problem """
    ydot[0] = y[1]
    ydot[1] = -g

def root_fn(t, y, out):
    """ root function to check the object reached height Y1 """
    out[0] = Y1 - y[0]
    return 0

def root_fn2(t, y, out):
    """ root function to check the object reached height Y1 """
    out[0] = Y1 - y[0]
    out[1] = (t-10)*(t-20)*(t-30)
    return 0

def root_fn3(t, y, out):
    """ root function to check the object reached time 10 """
    out[0] = (t - 10)*(t-20)*(t-30)
    return 0

def onroot_va(t, y, solver):
    """
    onroot function to reset the solver back at the start, but keep the current
    velocity
    """
    # Teleport the object back to height Y0, but retain its speed
    solver.reinit_IC(t, [Y0, y[1]])

    return 0

def onroot_vb(t, y, solver):
    """
    onroot function to stop solver when root is found
    """
    return 1

def onroot_vc(t, y, solver):
    """
    onroot function to reset the solver back at the start, but keep the current
    velocity as long as the time is less than a given amount
    """
    if t > 28: # we have found 4 interruption points, so we stop
        return 1
    solver.reinit_IC(t, [Y0, y[1]])
    return 0

def onroot_vd(t, y, solver):
    """
    onroot function to just continue if time <28
    """
    if t > 28:
        return 1

    return 0

n=0

def ontstop_va(t, y, solver):
    """
    ontstop function to reset the solver back at the start, but keep the current
    velocity
    """
    # Teleport the object back to height Y0, but retain its speed
    global n
    solver.reinit_IC(t, [Y0, y[1]])
    n += 1
    solver.set_options(tstop=T1+n*10)

    return 0

def ontstop_vb(t, y, solver):
    """
    ontstop function to stop solver when tstop is reached
    """
    return 1

def ontstop_vc(t, y, solver):
    """
    ontstop function to reset the solver back at the start, but keep the current
    velocity as long as the time is less than a given amount
    """
    global n
    if t > 28: # we have found 3 interruption points, so we stop
        return 1
    solver.reinit_IC(t, [Y0, y[1]])
    n += 1
    solver.set_options(tstop=T1+n*10)

    return 0

class TestOn(TestCase):
    """
    Check integrate.dae
    """

    def test_cvode_rootfn_noroot(self):
        #test calling sequence. End is reached before root is found
        tspan = np.arange(0, t_end1 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, nr_rootfns=1, rootfn=root_fn,
                     old_api=False)
        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.SUCCESS, "ERROR: Error occurred"
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [10.0, 509.4995, -98.10],
                        atol=atol, rtol=rtol)

    def test_cvode_rootfn(self):
        #test root finding and stopping: End is reached at a root
        tspan = np.arange(0, t_end2 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, nr_rootfns=1, rootfn=root_fn,
                     old_api=False)
        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.ROOT_RETURN, "ERROR: Root not found!"
        assert allclose([soln.roots.t[0], soln.roots.y[0,0], soln.roots.y[0,1]],
                        [14.206856, 10.0000, -139.3693],
                        atol=atol, rtol=rtol)

    def test_cvode_rootfnacc(self):
        #test root finding and accumilating: End is reached normally, roots stored
        tspan = np.arange(0, t_end2 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, nr_rootfns=1, rootfn=root_fn,
                     onroot=onroot_va,
                     old_api=False)
        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.SUCCESS, "ERROR: Error occurred"
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [100.0, 459.8927, -981.0000],
                        atol=atol, rtol=rtol)
        assert len(soln.roots.t) == 49, "ERROR: Did not find all 49 roots"
        assert allclose([soln.roots.t[-1], soln.roots.y[-1,0], soln.roots.y[-1,1]],
                        [99.447910, 10.0000, -975.5840],
                        atol=atol, rtol=rtol)

    def test_cvode_rootfn_stop(self):
        #test root finding and stopping: End is reached at a root with a function
        tspan = np.arange(0, t_end2 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, nr_rootfns=1, rootfn=root_fn,
                     onroot=onroot_vb,
                     old_api=False)
        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.ROOT_RETURN, "ERROR: Root not found!"
        assert allclose([soln.roots.t[-1], soln.roots.y[-1,0], soln.roots.y[-1,1]],
                        [14.206856, 10.0000, -139.3693],
                        atol=atol, rtol=rtol)

    def test_cvode_rootfn_test(self):
        #test root finding and accumilating: End is reached after a number of root
        tspan = np.arange(0, t_end2 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, nr_rootfns=1, rootfn=root_fn,
                     onroot=onroot_vc,
                     old_api=False)
        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.ROOT_RETURN, "ERROR: Not sufficient root found"
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [28.0, 124.4724, -274.6800],
                        atol=atol, rtol=rtol)
        assert len(soln.roots.t) == 4, "ERROR: Did not find all 4 roots"
        assert allclose([soln.roots.t[-1], soln.roots.y[-1,0], soln.roots.y[-1,1]],
                        [28.413692, 10.0000, -278.7383],
                        atol=atol, rtol=rtol)

    def test_cvode_rootfn_two(self):
        #test two root finding
        tspan = np.arange(0, t_end2 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, nr_rootfns=2, rootfn=root_fn2,
                     onroot=onroot_vc,
                     old_api=False)
        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.ROOT_RETURN, "ERROR: Not sufficient root found"
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [28.0, 106.4753, -274.6800],
                        atol=atol, rtol=rtol)
        assert len(soln.roots.t) == 5, "ERROR: Did not find all 4 roots"
        assert allclose([soln.roots.t[-1], soln.roots.y[-1,0], soln.roots.y[-1,1]],
                        [28.349052, 10.0000, -278.1042],
                        atol=atol, rtol=rtol)

    def test_cvode_rootfn_end(self):
        #test root finding with root at endtime
        tspan = np.arange(0, 30 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, nr_rootfns=1, rootfn=root_fn3,
                     onroot=onroot_vc,
                     old_api=False)
        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.ROOT_RETURN, "ERROR: Not sufficient root found"
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [30.0, -1452.5024, -294.3000],
                        atol=atol, rtol=rtol)
        assert len(soln.roots.t) == 3, "ERROR: Did not find all 4 roots"
        assert allclose([soln.roots.t[-1], soln.roots.y[-1,0], soln.roots.y[-1,1]],
                        [30.0, -1452.5024, -294.3000],
                        atol=atol, rtol=rtol)

    def test_cvode_tstopfn_notstop(self):
        #test calling sequence. End is reached before tstop is found
        global n
        n = 0
        tspan = np.arange(0, t_end1 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, tstop=T1+1, ontstop=ontstop_va,
                     old_api=False)

        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.SUCCESS, "ERROR: Error occurred"
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [10.0, 509.4995, -98.10],
                        atol=atol, rtol=rtol)

    def test_cvode_tstopfn(self):
        #test tstop finding and stopping: End is reached at a tstop
        global n
        n = 0
        tspan = np.arange(0, t_end2 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, tstop=T1,
                     old_api=False)
        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.TSTOP_RETURN, "ERROR: Tstop not found!"
        assert allclose([soln.tstop.t[0], soln.tstop.y[0,0], soln.tstop.y[0,1]],
                        [10.0, 509.4995, -98.10],
                        atol=atol, rtol=rtol)
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [10.0, 509.4995, -98.10],
                        atol=atol, rtol=rtol)

    def test_cvode_tstopfnacc(self):
        #test tstop finding and accumilating: End is reached normally, tstop stored
        global n
        n = 0
        tspan = np.arange(0, t_end2 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, tstop=T1, ontstop=ontstop_va,
                     old_api=False)
        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.SUCCESS, "ERROR: Error occurred"
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [100.0, -8319.5023, -981.00],
                        atol=atol, rtol=rtol)
        assert len(soln.tstop.t) == 9, "ERROR: Did not find all tstop"
        assert allclose([soln.tstop.t[-1], soln.tstop.y[-1,0], soln.tstop.y[-1,1]],
                        [90.0, -7338.5023, -882.90],
                        atol=atol, rtol=rtol)

    def test_cvode_tstopfn_stop(self):
        #test calling sequence. End is reached at a tstop
        global n
        n = 0
        tspan = np.arange(0, t_end2 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, tstop=T1, ontstop=ontstop_vb,
                     old_api=False)

        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.TSTOP_RETURN, "ERROR: Error occurred"
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [10.0, 509.4995, -98.10],
                        atol=atol, rtol=rtol)
        assert len(soln.tstop.t) == 1, "ERROR: Did not find all tstop"

        assert len(soln.values.t) == 11, "ERROR: Did not find all output"
        assert allclose([soln.tstop.t[-1], soln.tstop.y[-1,0], soln.tstop.y[-1,1]],
                        [10.0, 509.4995, -98.10],
                        atol=atol, rtol=rtol)

    def test_cvode_tstopfn_test(self):
        #test calling sequence. tstop function continues up to a time
        global n
        n = 0
        tspan = np.arange(0, t_end2 + 1, 1.0, DTYPE)
        solver = ode('cvode', rhs_fn, tstop=T1, ontstop=ontstop_vc,
                     old_api=False)

        soln = solver.solve(tspan, y0)
        assert soln.flag==StatusEnum.TSTOP_RETURN, "ERROR: Error occurred"
        assert allclose([soln.values.t[-1], soln.values.y[-1,0], soln.values.y[-1,1]],
                        [30.0, -1452.5024, -294.30],
                        atol=atol, rtol=rtol)
        assert len(soln.tstop.t) == 3, "ERROR: Did not find all tstop"
        assert len(soln.values.t) == 31, "ERROR: Did not find all output"
        assert allclose([soln.tstop.t[-1], soln.tstop.y[-1,0], soln.tstop.y[-1,1]],
                        [30.0, -1452.5024, -294.30],
                        atol=atol, rtol=rtol)
if __name__ == "__main__":
    try:
        run_module_suite()
    except NameError:
        test = TestOn()
        test.test_cvode_rootfn_noroot()
        test.test_cvode_rootfn()
        test.test_cvode_rootfnacc()
        test.test_cvode_rootfn_stop()
        test.test_cvode_rootfn_test()
        test.test_cvode_tstopfn()
        test.test_cvode_tstopfn_notstop()
        test.test_cvode_tstopfn_stop()
        test.test_cvode_tstopfn_test()
        test.test_cvode_tstopfnacc()
